# -*- coding: utf-8 -*-

import numpy as np
import numpy.linalg as la
import scipy.linalg as sla
# import scipy.sparse as sps
# import warnings
from .rsp import rsp, rsp_sa21Tlu, rsp_sb12Tlu


def s_1l(thickness, ql, al0, bl0):
    """
    calculate the scattering matrix of 1 layer.

    Parameters
    ----------
    thickness   :   float
                    thickness of layer
    ql          :   ndarray
                    1d array, eigen propagation constant in z in layer
    al0         :   ndarray
    bl0         :   ndarray
                    interface matrix

    Returns
    -------
    s11, s12, s21, s22  :   ndarray
                            four elements of the scattering matrix

    """

    a = al0
    b = bl0
    alu = sla.lu_factor(a)

    fl = np.exp(1j * ql * thickness)

    a_fbafb = a - fl[:, None] * b @ sla.lu_solve(alu, (fl[:, None] * b))
    a_fbafb_lu = sla.lu_factor(a_fbafb)

    fbafa_b = (fl[:, None] * b @ sla.lu_solve(alu, fl[:, None] * a) - b)
    fa_bab = (fl[:, None] * (a - b @ sla.lu_solve(alu, b)))

    s11 = sla.lu_solve(a_fbafb_lu, fbafa_b)
    s12 = sla.lu_solve(a_fbafb_lu, fa_bab)

    s22 = s11
    s21 = s12

    return s11, s12, s21, s22


def s_1l_rsp(thickness: float,
             ql: np.ndarray,
             afl: np.ndarray,
             bfl: np.ndarray):
    """
    Calculate the scattering matrix of 1 layer, using Redheffer products, using interface matrices from outside to this layer. The layer is assumed to be sandwiched between two infinitely thin layer of some other materials.

    Parameters
    ----------
    thickness : thickness of layer
    ql : 1d array, eigen propagation constant in z in layer
    afl : interface matrix
    bfl : interface matrix

    Returns
    -------

    """
    a = afl
    b = bfl

    tng = afl.shape[0]

    f = np.exp(1j * ql * thickness)

    # ia = la.inv(a)
    aTlu = sla.lu_factor(a.T)
    aTlu2 = (aTlu[0].copy(), aTlu[1].copy())
    a1 = aTlu2[0]
    a1[np.triu_indices(a1.shape[0])] *= 0.5
    alu = sla.lu_factor(a)
    # alu2 = (alu[0].copy(), alu[1].copy())
    # a1 = alu2[0]
    # a1[np.triu_indices(a1.shape[0])] *= 0.5

    ab = sla.lu_solve(alu, b)
    ba = sla.lu_solve(aTlu, b.T).T
    bfa = sla.lu_solve(aTlu, (b*f).T).T

    bab = b @ ab
    a_bab = a - bab
    a_babf = a_bab * f
    a_bfabf = a - b * f @ ab * f
    term = sla.solve(a_bfabf.T, a_babf.T).T

    s11 = ba - term @ bfa
    s12 = term
    s21 = s12
    s22 = s11

    # alu = sla.lu_factor(a)
    # aib = sla.lu_solve(alu, b)
    # sl11 = b @ ia
    sl11 = sla.lu_solve(aTlu, b.T).T
    # sl12 = a - b @ ia @ b
    sl12 = 0.5 * (a - b @ ab)
    sl21 = aTlu2
    sl22 = -ab

    # # debugging
    # ts = np.ones(tng, dtype=complex)
    # ts[2] *= 1e7
    # ts[4] *= 1e9
    #
    # term_ts = sla.solve((a_bfabf*ts).T, (a_babf*ts).T).T
    # diff = term - term_ts
    # diffmag = np.abs(diff).max()
    # # end of debugging


    # # explicit rsp subroutine
    # sf11 = np.zeros((tng, tng))
    # sf12 = np.diag(f)
    # sf21 = np.diag(f)
    # sf22 = np.zeros((tng, tng))
    #
    # sr11 = sl22
    # sr12 = sl21
    # sr21 = sl12
    # sr22 = sl11
    #
    # ss = rsp_sa21Tlu(sl11, sl12, sl21, sl22, sf11, sf12, sf21, sf22)
    # sm = rsp_sb12Tlu(*ss, sr11, sr12, sr21, sr22)

    # # debugging
    # f = np.exp(1j * ql * 0.35)
    # sngl1 = (a - b @ ab)*f
    # sngl2 = a - b * f @ ab * f
    # prdc = sla.solve(sngl2.T, sngl1.T).T
    # prdc1 = prdc.copy()
    # prdc1[np.where(np.abs(prdc) < 1e-14)] = 0.
    # prdctest = sla.solve(sngl1.T, sngl1.T).T

    # # debugging
    # # wong left/right scattering matrix
    # f = np.exp(1j * ql * 0.5)
    # sngl1 = f * (a - b @ ab)
    # sngl2 = sla.solve(b.T, a).T - f * sla.solve(a.T, b).T * f
    # prdc = sla.solve(sngl2, sngl1)
    # prdc1 = prdc.copy()
    # prdc1[np.where(np.abs(prdc) < 1e-14)] = 0.
    # # prdctest = sla.solve(sngl1.T, sngl1.T).T

    return s11, s12, s21, s22


def s_1l_rsp_lrd(thickness: float,
             ql: np.ndarray,
             afl: np.ndarray,
             bfl: np.ndarray,
             afl2: np.ndarray,
             bfl2: np.ndarray):
    """
    Calculate the scattering matrix of 1 layer, using Redheffer products, using interface matrices from outside to this layer. The layer is assumed to be sandwiched between two infinitely thin layer of some other materials.

    lrd: assuming left and right different

    Parameters
    ----------
    thickness : thickness of layer
    ql : 1d array, eigen propagation constant in z in layer
    afl : interface matrix
    bfl : interface matrix

    Returns
    -------

    """
    a = afl
    b = bfl

    tng = afl.shape[0]

    # ia = la.inv(a)
    aTlu = sla.lu_factor(a.T)
    aTlu2 = (aTlu[0].copy(), aTlu[1].copy())
    a1 = aTlu2[0]
    a1[np.triu_indices(a1.shape[0])] *= 0.5
    alu = sla.lu_factor(a)
    # alu2 = (alu[0].copy(), alu[1].copy())
    # a1 = alu2[0]
    # a1[np.triu_indices(a1.shape[0])] *= 0.5

    ab = sla.lu_solve(alu, b)

    # alu = sla.lu_factor(a)
    # aib = sla.lu_solve(alu, b)
    # sl11 = b @ ia
    sl11 = sla.lu_solve(aTlu, b.T).T
    # sl12 = a - b @ ia @ b
    sl12 = 0.5 * (a - b @ ab)
    sl21 = aTlu2
    sl22 = -ab

    f = np.exp(1j * ql * thickness)

    sf11 = np.zeros((tng, tng))
    sf12 = np.diag(f)
    sf21 = np.diag(f)
    sf22 = np.zeros((tng, tng))

    a2 = afl2
    b2 = bfl2

    # ia = la.inv(a)
    a2Tlu = sla.lu_factor(a2.T)
    a2Tlu2 = (a2Tlu[0].copy(), a2Tlu[1].copy())
    a1 = a2Tlu2[0]
    a1[np.triu_indices(a1.shape[0])] *= 0.5
    a2lu = sla.lu_factor(a2)
    # alu2 = (alu[0].copy(), alu[1].copy())
    # a1 = alu2[0]
    # a1[np.triu_indices(a1.shape[0])] *= 0.5

    a2b2 = sla.lu_solve(a2lu, b2)

    sr11 = -a2b2
    sr12 = a2Tlu2
    sr21 = 0.5 * (a2 - b2 @ a2b2)
    sr22 = sla.lu_solve(a2Tlu, b2.T).T

    # # debugging
    # f = np.exp(1j * ql * 0.35)
    # sngl1 = (a - b @ ab)*f
    # sngl2 = a - b * f @ ab * f
    # prdc = sla.solve(sngl2.T, sngl1.T).T
    # prdc1 = prdc.copy()
    # prdc1[np.where(np.abs(prdc) < 1e-14)] = 0.
    # prdctest = sla.solve(sngl1.T, sngl1.T).T

    # # debugging
    # # wong left/right scattering matrix
    # f = np.exp(1j * ql * 0.5)
    # sngl1 = f * (a - b @ ab)
    # sngl2 = sla.solve(b.T, a).T - f * sla.solve(a.T, b).T * f
    # prdc = sla.solve(sngl2, sngl1)
    # prdc1 = prdc.copy()
    # prdc1[np.where(np.abs(prdc) < 1e-14)] = 0.
    # # prdctest = sla.solve(sngl1.T, sngl1.T).T

    ss = rsp_sa21Tlu(sl11, sl12, sl21, sl22, sf11, sf12, sf21, sf22)
    sm = rsp_sb12Tlu(*ss, sr11, sr12, sr21, sr22)

    return sm


def s_1l_1212(a, b):
    """
    Calculate the scattering matrix of material 1 to 2 using interface matrices defined from 1 to 2.

    Parameters
    ----------
    a     :   ndarray
    b     :   ndarray

    Returns
    -------
    s11  :  ndarray
    s12  :  ndarray
    s21  :  Tuple[np.ndarray, np.ndarray]
            lu factorization
    s22  :  ndarray
    """

    a = a
    b = b

    aTlu = sla.lu_factor(a.T)
    aTlu2 = (aTlu[0].copy(), aTlu[1].copy())
    aT1 = aTlu2[0]
    aT1[np.triu_indices(aT1.shape[0])] *= 0.5
    alu = sla.lu_factor(a)
    alu2 = (alu[0].copy(), alu[1].copy())
    a1 = alu2[0]
    a1[np.triu_indices(a1.shape[0])] *= 0.5
    ab = sla.lu_solve(alu, b)

    s11 = sla.lu_solve(aTlu, b.T).T
    s12 = 1./2. * (a - b @ ab)
    s21 = alu2
    s22 = - ab

    return s11, s12, s21, s22


def s_1l_1221(a, b):
    """
    Calculate the scattering matrix of the interface from material 1 to 2 using interface matrices defined from 2 to 1.

    Parameters
    ----------
    a     :   ndarray
    b     :   ndarray

    Returns
    -------
    s11  :  ndarray
    s12  :  Tuple[np.ndarray, np.ndarray]
            lu factorization
    s21  :  ndarray
    s22  :  ndarray
    """

    a = a
    b = b

    alu = sla.lu_factor(a)
    alu2 = (alu[0].copy(), alu[1].copy())
    a1 = alu2[0]
    a1[np.triu_indices(a1.shape[0])] *= 0.5
    aTlu = sla.lu_factor(a.T)
    aTlu2 = (aTlu[0].copy(), aTlu[1].copy())
    at1 = aTlu2[0]
    at1[np.triu_indices(at1.shape[0])] *= 0.5
    ab = sla.lu_solve(alu, b)

    s11 = - ab
    s12 = alu2
    s21 = 1./2. * (a - b @ ab)
    s22 = sla.lu_solve(aTlu, b.T).T

    return s11, s12, s21, s22





from inkstone.params import Params
from inkstone.mtr import Mtr
from inkstone.bx import Bx
import time
import numpy as np
import matplotlib.pyplot as plt

mtr = Mtr(2, 3, 'fic')

lv = ((3., 0), (2., 3))
pr = Params(latt_vec=lv)

a = 1.5
b = 0.8
cen = (0.2, 0.3)
angle = 30

bx = Bx(mtr, 'ellipse', center=cen, half_lengths=(a, b), angle=angle)

m = 50
n = 50
k1i = np.arange(-m, m + 1)
k2i = np.arange(-n, n + 1)
kk1i, kk2i = np.meshgrid(k1i, k2i)
b1, b2 = pr.recipr_vec
b1 = np.array(b1)
b2 = np.array(b2)
kk1, kk2 = [kk1i * b1[i] + kk2i * b2[i] for i in range(2)]
ks = list(zip(kk1.ravel(), kk2.ravel()))

t1 = time.process_time()
ep, ei, mu, mi = bx.ft(ks, factor=0.7)
t2 = time.process_time()
s = np.array(ep).reshape((2*n+1, 2*m+1, 3, 3))[:, :, 0, 0]
t3 = time.process_time()
print(t2-t1)
print(t3-t2)

xi = np.linspace(-0.5, 0.5, 101)
yi = np.linspace(-0.5, 0.5, 101)
xxi, yyi = np.meshgrid(xi, yi)
xx, yy = [xxi * lv[0][i] + yyi * lv[1][i] for i in range(2)]
# s = ly.epsi_fs[:, :, 0, 0]  # epsi_xx

poly_recon = 1. / 4 / np.pi ** 2 * np.cross(pr.recipr_vec[0], pr.recipr_vec[1]) * \
             np.sum(s[:, :, None, None] * np.exp(1j * ((kk1i * b1[0] + kk2i * b2[0])[:, :, None, None] * xx[None, None, :, :] + (kk1i * b1[1] + kk2i * b2[1])[:, :, None, None] * yy[None, None, :, :])), axis=(0, 1))

fig, ax = plt.subplots()
pcm = ax.pcolormesh(xx, yy, poly_recon.real)
ax.set_aspect('equal')
plt.colorbar(pcm)

plt.figure()
plt.plot(xxi[21, :], poly_recon.real[:, 21])



# -*- coding: utf-8 -*-

from typing import Tuple, Union, Optional
import numpy as np
# import numpy.linalg as la
# from warnings import warn


class Mtr:
    # material need to have a name such that user can access it by its name
    def __init__(self,
                 epsi: Union[Union[float, complex], Tuple[Union[float, complex], Union[float, complex], Union[float, complex]], np.ndarray],
                 mu: Union[Union[float, complex], Tuple[Union[float, complex], Union[float, complex], Union[float, complex]], np.ndarray],
                 name=None):
        """
        Material.
        """
        self._epsi: Optional[np.ndarray] = None
        self._mu: Optional[np.ndarray] = None
        self._epsi_inv: Optional[np.ndarray] = None
        self._mu_inv: Optional[np.ndarray] = None

        self._ep_is_vac = True
        self._mu_is_vac = True
        self.is_vac: bool = True

        self.is_diagonal: bool = True
        self.is_isotropic: bool = True
        self.is_dege: bool = True

        self.ep_is_diagonal: bool = True
        self.ep_is_isotropic: bool = True

        self.mu_is_diagonal: bool = True
        self.mu_is_isotropic: bool = True

        self.epsi = epsi
        self.mu = mu

        self.name = name

    @property
    def ep_is_diagonal(self) -> bool:
        return self._ep_is_diagonal

    @ep_is_diagonal.setter
    def ep_is_diagonal(self, val):
        self._ep_is_diagonal = val
        self._check_diagonal()
        self._check_isotropic()
        self._check_degenerate()

    @property
    def ep_is_isotropic(self) -> bool:
        return self._ep_is_isotropic

    @ep_is_isotropic.setter
    def ep_is_isotropic(self, val):
        self._ep_is_isotropic = val
        self._check_diagonal()
        self._check_isotropic()
        self._check_degenerate()

    @property
    def ep_is_vac(self) -> bool:
        return self._ep_is_vac

    @ep_is_vac.setter
    def ep_is_vac(self, val):
        self._ep_is_vac = val
        self._check_vac()

    @property
    def mu_is_diagonal(self) -> bool:
        return self._mu_is_diagonal

    @mu_is_diagonal.setter
    def mu_is_diagonal(self, val):
        self._mu_is_diagonal = val
        self._check_diagonal()
        self._check_isotropic()
        self._check_degenerate()

    @property
    def mu_is_isotropic(self) -> bool:
        return self._mu_is_isotropic

    @mu_is_isotropic.setter
    def mu_is_isotropic(self, val):
        self._mu_is_isotropic = val
        self._check_diagonal()
        self._check_isotropic()
        self._check_degenerate()

    @property
    def mu_is_vac(self) -> bool:
        return self._mu_is_vac

    @mu_is_vac.setter
    def mu_is_vac(self, val):
        self._mu_is_vac = val
        self._check_vac()

    @property
    def epsi(self) -> np.ndarray:
        return self._epsi

    @epsi.setter
    def epsi(self, val):
        if (type(val) is float) or (type(val) is int) or (type(val) is complex) or (type(val) is np.float64) or (type(val) is np.complex128):
            ep = np.eye(3, dtype=complex) * val
            self._epsi = ep + 0j
            self.ep_is_diagonal = True
            self.ep_is_isotropic = True
            if val == 1.:
                self.ep_is_vac = True
            else:
                self.ep_is_vac = False
        elif np.array(val).ndim == 1 and np.array(val).size == 3:
            ep = np.diag(val) + 0j
            self._epsi = ep + 0j
            self.ep_is_diagonal = True
            if check_iso(ep):
                self.ep_is_isotropic = True
            else:
                self.ep_is_isotropic = False
            if val[0] == 1. and val[1] == 1. and val[2] == 1.:
                self.ep_is_vac = True
            else:
                self.ep_is_vac = False
        elif (type(val) is not np.ndarray) or val.shape != (3, 3):
            raise ValueError('data format for epsilon is incorrect.')
        else:
            ep = val
            self._epsi = val + 0j
            if check_diag(val):
                self.ep_is_diagonal = True
                if check_iso(val):
                    self.ep_is_isotropic = True
                    if val[0, 0] == 1. and val[1, 1] == 1. and val[2, 2] == 1.:
                        self.ep_is_vac = True
                    else:
                        self.ep_is_vac = False
                else:
                    self.ep_is_isotropic = False
                    self.ep_is_vac = False
            else:
                self.ep_is_diagonal = False
                self.ep_is_isotropic = False
                self.ep_is_vac = False

        # explicit computation of inverse.
        v = ep
        adbc = v[0, 0] * v[1, 1] - v[0, 1] * v[1, 0]
        if adbc == 0 or v[2, 2] == 0:
            raise Exception('Singular permittivity tensor.')
        ei = np.array([[v[1, 1] / adbc, -v[0, 1] / adbc, 0],
                       [-v[1, 0] / adbc, v[0, 0] / adbc, 0],
                       [0, 0, 1 / v[2, 2]]], dtype=complex)
        self._epsi_inv = ei

    @property
    def mu(self) -> np.ndarray:
        return self._mu

    @mu.setter
    def mu(self, val):
        if (type(val) is float) or (type(val) is int) or (type(val) is complex) or (type(val) is np.float64) or (type(val) is np.complex128):
            mu = np.eye(3, dtype=complex) * val
            self._mu = mu + 0j
            self.mu_is_diagonal = True
            self.mu_is_isotropic = True
            if val == 1.:
                self.mu_is_vac = True
            else:
                self.mu_is_vac = False
        elif np.array(val).ndim == 1 and np.array(val).size == 3:
            mu = np.diag(val) + 0j
            self._mu = mu + 0j
            self.mu_is_diagonal = True
            if check_iso(mu):
                self.mu_is_isotropic = True
            else:
                self.mu_is_isotropic = False
            if val[0] == 1. and val[1] == 1. and val[2] == 1.:
                self.mu_is_vac = True
            else:
                self.mu_is_vac = False
        elif (type(val) is not np.ndarray) or val.shape != (3, 3):
            raise ValueError('data format for mu is incorrect.')
        else:
            mu = val
            self._mu = val + 0j
            if check_diag(val):
                self.mu_is_diagonal = True
                if check_iso(val):
                    self.mu_is_isotropic = True
                    if val[0, 0] == 1. and val[1, 1] == 1. and val[2, 2] == 1.:
                        self.mu_is_vac = True
                    else:
                        self.mu_is_vac = False
                else:
                    self.mu_is_isotropic = False
                    self.mu_is_vac = False
            else:
                self.mu_is_diagonal = False
                self.mu_is_isotropic = False
                self.mu_is_vac = False

        # explicit computation of inverse.
        v = mu
        adbc = v[0, 0] * v[1, 1] - v[0, 1] * v[1, 0]
        if adbc == 0 or v[2, 2] == 0:
            raise Exception('Singular permittivity tensor.')
        mi = np.array([[v[1, 1] / adbc, -v[0, 1] / adbc, 0],
                       [-v[1, 0] / adbc, v[0, 0] / adbc, 0],
                       [0, 0, 1 / v[2, 2]]], dtype=complex)
        self._mu_inv = mi

    @property
    def epsi_inv(self) -> np.ndarray:
        return self._epsi_inv

    @property
    def mu_inv(self) -> np.ndarray:
        return self._mu_inv

    def _check_isotropic(self):
        """check if material is isotropic."""
        if (self.epsi is not None) and (self.mu is not None):
            if self.ep_is_isotropic and self.mu_is_isotropic:
                self.is_isotropic = True
            else:
                self.is_isotropic = False

            # ep = self.epsi
            # mu = self.mu
            # if check_iso(ep) and check_iso(mu):
            #     self.is_isotropic = True
            # else:
            #     self.is_isotropic = False

    def _check_diagonal(self):
        """check if material is diagonal."""
        if (self.epsi is not None) and (self.mu is not None):
            if self.ep_is_diagonal and self.mu_is_diagonal:
                self.is_diagonal = True
            else:
                self.is_diagonal = False

            # ep = self.epsi
            # mu = self.mu
            # if check_diag(ep) and check_diag(mu):
            #     self.is_diagonal = True
            # else:
            #     self.is_isotropic = False

    def _check_degenerate(self):
        """check if material is degenerate eigen"""
        if (self.epsi is not None) and (self.mu is not None):
            if self.ep_is_diagonal and self.mu_is_diagonal:
                if self.ep_is_isotropic and self.mu_is_isotropic:
                    self.is_dege = True
                else:
                    ep = self.epsi
                    mu = self.mu
                    if check_dege(ep, mu):
                        self.is_dege = True
                    else:
                        self.is_dege = False
            else:
                self.is_dege = False

    def _check_vac(self):
        if self.ep_is_vac and self.mu_is_vac:
            self.is_vac = True
        else:
            self.is_vac = False

def check_iso(val):
    """check of 3x3 tensor is isotropic"""
    if check_diag(val):
        if (val[0, 0] == val[1, 1]) and (val[0, 0] == val[2, 2]):
            return True
        else:
            return False
    else:
        return False


def check_diag(val):
    """check of 3x3 tensor is diagonal"""
    if (val[0, 1] == 0.) and (val[0, 2] == 0.) and (val[1, 0] == 0.) and (val[1, 2] == 0.) and (val[2, 0] == 0.) and (val[2, 1] == 0.):
        return True
    else:
        return False


def check_dege(ep, mu):
    """check if epsi and mu 3x3 tensors satisfy degenerate condition"""
    if check_diag(ep) and check_diag(mu):
        mxx, mxy, mxz, myx, myy, myz, mzx, mzy, mzz = mu.ravel()
        exx, exy, exz, eyx, eyy, eyz, ezx, ezy, ezz = ep.ravel()
        if (eyy * mzz == ezz * myy) and (exx * mzz == ezz * mxx):
            return True
        else:
            return False
    else:
        return False



# -*- coding: utf-8 -*-

import numpy as np
# from numpy import linalg as la


def g_pts_1d(num_g, b):
    """
    given number of lattice points and one 2D lattice vector, return the list of lattice points and the index of lattice points.

    Parameters
    ----------
    num_g       :   int
                    target total number of lattice points
    b           :   tuple[float, float]
                    lattice vector

    Returns
    -------
    k_pts       :   list[tuple[float, float]]
                    the k points
    idx         :   list[int]
                    the indices of the k points, i.e. k = idx * b

    """

    b = np.array(b)

    imax = int(num_g / 2)

    idx = [i - 2 * i * j for i in range(imax+1) for j in range(2)][1:]
    # This gives [0, 1, -1, 2, -2, ..., imax, -imax]

    k_pts = [(b[0] * i, b[1] * i) for i in idx]

    return k_pts, idx




# -*- coding: utf-8 -*-

import numpy as np
from inkstone.g_pts import g_pts

import matplotlib.pyplot as plt

num_g = 50
b1 = (1., 0.)
b2 = (0.8, 0.9)

k_pts, idx = g_pts(num_g, b1, b2)
ka = np.array(k_pts)
idxa = np.array(idx)

fig = plt.figure()
ax = fig.add_subplot(111)
ax.scatter(ka[:, 0], ka[:, 1])
ax.set_aspect('equal')

fig1 = plt.figure()
ax1 = fig1.add_subplot(111)
ax1.scatter(idxa[:, 0], idxa[:, 1])
ax1.set_aspect('equal')



# -*- coding: utf-8 -*-

import numpy as np
import numpy.linalg as la
from scipy.special import jn


def ft_2d_ellip(a, b, ks, center=None, angle=0.):
    """
    Calculate the fourier transform of a function with value 1 inside a ellipse and 0 outside.

    The ellipse has half widths a and b, rotated ccw of `angle`, then shifted to `center`.

    Parameters
    ----------
    a, b        :   float
                    half widths of the ellipse, assume a in x direction and b in y
    ks          :   list[tuple[float, float]]
                    list of (kx, ky) points
    center      :   tuple[float, float]
                    the center of the disk
    angle       :   float
                    the rotation angle of the ellipse, ccw, in degrees

    Returns
    -------
    s       :   list[complex]
                1d array, Fourier coefficient at the input ks positions

    """

    ang = angle / 180. * np.pi
    stretch = np.array([[a, 0], [0, b]])
    rotate = np.array([[np.cos(ang), -np.sin(ang)], [np.sin(ang), np.cos(ang)]])
    aff = rotate @ stretch

    ksa = np.array(ks)  # nx2 shape
    aksa = (aff.T @ ksa.T).T
    aks_nm = np.linalg.norm(aksa, axis=-1)  # 1d array of n. The norm of each k vector
    idx_0 = np.where(aks_nm == 0)[0]  # index to where k is (0, 0)
    idx_i = np.where(aks_nm != 0)[0]  # index to where k is not (0, 0)
    aks_nm1 = np.delete(aks_nm, idx_0)

    ksa1 = np.delete(ksa, idx_0, axis=0)  # new ks array that doesn't contain (0, 0)

    if center is None:
        center = (0, 0)
    cent = np.array(center)

    s = 1j * np.zeros(aks_nm.size)
    s[idx_i] = np.abs(la.det(aff)) * 2 * np.pi * jn(1, aks_nm1) / aks_nm1 * np.exp(-1j * cent @ ksa1.T)
    s[idx_0] = np.pi * a * b

    return s.tolist()




# -*- coding: utf-8 -*-

import numpy as np
from warnings import warn
import copy
from typing import List, Union, Tuple, Optional
from inkstone.ft.ft_1d_sq import ft_1d_sq
from inkstone.ft.ft_2d_rct import ft_2d_rct
from inkstone.ft.ft_2d_para import ft_2d_para
from inkstone.ft.ft_2d_ellip import ft_2d_ellip
from inkstone.ft.ft_2d_disk import ft_2d_disk
from inkstone.ft.ft_2d_poly import ft_2d_poly
from inkstone.ft.poly_area import poly_area
from inkstone.ft.gibbs import gibbs_corr


class Shp:
    # Shp doesn't know what lattice it is in
    # Shape recalculate its fourier series at read time (when accessing self.ft()).
    # FT calculation is analytic hence quick, recalculate every time ft() is called. No need to check if parameters modified

    def __init__(self,
                 shp: str,
                 ks: List[Union[float, Tuple[float, float]]] =None,
                 **kw_gibbs):
        """
        Basic shape.

        You can specify the geometry of the shape, and get the Fourier transform (FT) of the shape. The FT of a shape is understood as a function that is 1 inside and 0 outside.

        You can choose whether to have Gibbs correction for the returned FT coefficients.

        You can give the list of k points `ks` at initialization time, use `Shp.ks` to set it, or supply it at `Shp.ft(ks)`. with `ks` given, you can call `Shp.ft()` to get the Fourier transform on the k points.

        Parameters
        ----------
        shp                 :   what kind of shape this is
        ks                  :   list of k points to calculate FT on
        kw_gibbs            :   keyword arguments for Gibbs correction, to be passed to `.ft.gibbs.gibbs_corr`
        """
        self.shape: str = shp
        self.area: Optional[float] = None
        self._ft: Optional[List[float]] = None  # 1d array Fourier series at the k points
        self._if_gibbs_corr = True
        self._kw_gibbs: dict = {}

        self.ks = ks  # list of k points to calculate Fourier series
        self.use_gibbs_correction(**kw_gibbs)

    @property
    def ks(self) -> List[Union[float, Tuple[float, float]]]:
        """The list of 1d k points or 2d (kx, ky) tuples on which the Fourier coefficients of the shape is calculated."""
        return self._ks

    @ks.setter
    def ks(self, val: List[Union[float, Tuple[float, float]]]):
        self._ks: List[Union[float, Tuple[float, float]]] = val  # don't do deep copy which would take longer time than the actual Fourier transform.

    def use_gibbs_correction(self, if_gibbs_correction: bool = None, **kw_gibbs):
        """
        Choose whether to have correction for Gibbs phenomenon.
        Parameters
        ----------
        if_gibbs_correction :   whether to use gibbs correction when calculating the Fourier transform
        kw_gibbs            :   other keyword arguments for Gibbs correction, to be passed on to `.ft.gibbs.gibbs_corr`
        -------

        """
        if if_gibbs_correction is not None:
            self._if_gibbs_corr = if_gibbs_correction
        for key, value in kw_gibbs.items():
            self._kw_gibbs[key] = value

    def ft(self,
           ks: Optional[List[Tuple[float, float]]] = None,
           **kw_gibbs
           ) -> Optional[List[complex]]:
        """
        Get the Fourier transform coefficients on the given k points in `ks`.
        Can choose to have correction for Gibbs phenomenon.

        Parameters
        ----------
        ks          :   list of k points to calculate FT on
        kw_gibbs    :   keyword arguments for Gibbs correction, to be passed on to `.ft.gibbs.gibbs_corr`.
        """
        if ks:
            self.ks: List = ks

        self._ft = self._calc_ft()
        s = 1
        self.use_gibbs_correction(**kw_gibbs)
        if self._if_gibbs_corr:
            s = gibbs_corr(self.ks, **self._kw_gibbs)
        ft = (np.array(self._ft, dtype=complex) * np.array(s, dtype=complex)).tolist()
        self._ft = ft
        return self._ft

    def _calc_ft(self) -> List[complex]:
        """
        Get fourier coefficients at given k points
        To be implemented in different shapes.
        """

    use_gibbs_correction.__doc__ += gibbs_corr.__doc__
    __init__.__doc__ += use_gibbs_correction.__doc__
    __init__.__doc__ += ks.__doc__
    ft.__doc__ += ks.__doc__
    ft.__doc__ += use_gibbs_correction.__doc__


class OneD(Shp):
    def __init__(self, width: float, center: float = None, **kwargs):
        """
        This is a 1D shape.

        If you query the Fourier transform coefficients at a list of 2D k points (kx, ky), the first nonzero k components will be retained while the other be ignored. For example, [(0, 1), (2, 3)] effectively become [(0, 1), (0, 3)]

        The area of this shape is defined as its width.
        """
        super(OneD, self).__init__('1d', **kwargs)

        self._center = None

        self.width = width
        self.center = center

    @property
    def width(self) -> float:
        """The width of the 1d shape."""
        return self._width

    @width.setter
    def width(self, val: float):
        self._width = val
        self.area = val

    @property
    def center(self) -> float:
        """The center of the 1d shape"""
        return self._center

    @center.setter
    def center(self, val: float):
        if val is None:
            if self._center is None:
                self._center = 0.
        else:
            self._center = val

    @property
    def ks(self) -> List[float]:
        """The list of 1d k points or 2d (kx, ky) tuples on which the Fourier coefficients of the shape is calculated."""
        return self._ks

    @ks.setter
    def ks(self, val: List[Union[float, Tuple[float, float]]]):
        if val:
            if type(val[0]) is tuple:
                for v in val:
                    if v[0] != 0:
                        self._ks: List[float] = [k[0] for k in val]
                        break
                    elif v[1] != 0:
                        self._ks: List[float] = [k[1] for k in val]
                        break
                self._ks: List[float] = [k[0] for k in val]
            else:
                self._ks: List[float] = val

    def _calc_ft(self) -> List[complex]:
        return ft_1d_sq(self.width, self.ks, self.center)


class Rect(Shp):
    def __init__(self, side_lengths, center=None, angle=None, **kwargs):
        """
        A rectangular shape.

        Parameters
        ----------
        side_lengths    :   tuple[float, float]
        center          :   tuple[float, float]
        angle           :   float
        """

        super(Rect, self).__init__('rectangle', **kwargs)

        self._center = None
        self._angle = None

        self.side_lengths = side_lengths
        self.center = center
        self.angle = angle

    @property
    def side_lengths(self):
        return self._side_lengths

    @side_lengths.setter
    def side_lengths(self, val):
        # assume when setting side lengths it must have been changed. no `If` which is slow.
        self.area = val[0] * val[1]
        self._side_lengths = val

    @property
    def center(self):
        return self._center

    @center.setter
    def center(self, val):
        if val is None:
            if self._center is None:
                self._center = (0., 0.)
        else:
            self._center = val

    @property
    def angle(self):
        return self._angle

    @angle.setter
    def angle(self, val):
        if val is None:
            if self._angle is None:
                self._angle = 0.
        else:
            self._angle = val

    def _calc_ft(self) -> List[complex]:
        """
        Get fourier coefficients at given k points
        """
        return ft_2d_rct(self.side_lengths[0], self.side_lengths[1], self.ks, self.center, self.angle)


class Para(Shp):
    def __init__(self, side_lengths, center=None, shear_angle=None, angle=None, **kwargs):
        """
        A parallelogram shape.

        This shape has two actual side lengths defined here.

        Can define the shear angle, i.e. the included angle between the two arms.

        Then the parallelogram can be rotated by `angle` and then shifted to new `center`.

        Parameters
        ----------
        side_lengths    :   tuple[float, float]
                            actual side lengths of the shape
        center          :   tuple[float, float]
        shear_angle     :   float
                            the angle between the two side arms in degrees.
        angle           :   float
                            the rotation angle of the entire shape in degrees

        """

        super(Para, self).__init__('parallelogram', **kwargs)

        self.side_lengths = side_lengths

        self._center = None
        self._shear_angle = None
        self._angle = None

        self.center = center
        self.angle = angle
        self.shear_angle = shear_angle

    @property
    def side_lengths(self):
        return self._side_lengths

    @side_lengths.setter
    def side_lengths(self, val):
        # assume when setting side lengths it must have been changed. no `If` which is slow.
        self.area = val[0] * val[1]
        self._side_lengths = val

    @property
    def center(self):
        return self._center

    @center.setter
    def center(self, val):
        if val is None:
            if self._center is None:
                self._center = (0., 0.)
        else:
            self._center = val

    @property
    def angle(self):
        return self._angle

    @angle.setter
    def angle(self, val):
        if val is None:
            if self._angle is None:
                self._angle = 0.
        else:
            self._angle = val

    @property
    def shear_angle(self):
        return self._shear_angle

    @shear_angle.setter
    def shear_angle(self, val):
        if val is None:
            if self._shear_angle is None:
                self._shear_angle = 0.
        else:
            self._shear_angle = val

    def _calc_ft(self) -> List[complex]:
        """
        Get fourier coefficients at given k points
        """
        return ft_2d_para(a=self.side_lengths[0], b=self.side_lengths[1], ks=self.ks, center=self.center, shear_angle=self.shear_angle, rotate_angle=self.angle)


class Elli(Shp):
    def __init__(self, half_lengths, center=None, angle=None, **kwargs):
        """
        A ellipse shape.

        Parameters
        ----------
        half_lengths    :   tuple[float, float]
        center          :   tuple[float, float]
        angle           :   float
        """

        super(Elli, self).__init__('ellipse', **kwargs)

        self.half_lengths = half_lengths

        self._center = None
        self._angle = None

        self.center = center
        self.angle = angle

    @property
    def half_lengths(self):
        return self._half_lengths

    @half_lengths.setter
    def half_lengths(self, val):
        # assume when setting side lengths it must have been changed. no `If` which is slow.
        self._half_lengths = val
        self.area = np.pi * val[0] * val[1]

    @property
    def center(self):
        return self._center

    @center.setter
    def center(self, val):
        if val is None:
            if self._center is None:
                self._center = (0., 0.)
        else:
            self._center = val

    @property
    def angle(self):
        return self._angle

    @angle.setter
    def angle(self, val):
        if val is None:
            if self._angle is None:
                self._angle = 0.
        else:
            self._angle = val

    def _calc_ft(self) -> List[complex]:
        """
        Get fourier coefficients at given k points
        """
        return ft_2d_ellip(self.half_lengths[0], self.half_lengths[1], self.ks, self.center, self.angle)


class Disk(Shp):

    def __init__(self, radius, center=None, **kwargs):
        """
        A disk shape.
        """
        super(Disk, self).__init__('disk', **kwargs)

        self.radius = radius

        self._center = None

        self.center = center

    @property
    def radius(self):
        """Return the radius of the disk."""
        return self._radius

    @radius.setter
    def radius(self, val):
        # assume when setting side lengths it must have been changed. no `If` which is slow.
        self._radius = val
        self.area = np.pi * val ** 2

    @property
    def center(self):
        """Return the center of the disk."""
        return self._center

    @center.setter
    def center(self, val):
        if val is None:
            if self._center is None:
                self._center = (0., 0.)
        else:
            self._center = val

    def _calc_ft(self) -> List[complex]:
        """
        Get fourier coefficients at given k points
        """
        return ft_2d_disk(self.radius, self.ks, self.center)


class Poly(Shp):
    def __init__(self, vertices, **kwargs):
        """
        A polygon shape

        Parameters
        ----------
        vertices        :   list[tuple[float, float]]
        """

        super(Poly, self).__init__('polygon', **kwargs)

        self.vertices = vertices

    @property
    def vertices(self):
        return self._vertices

    @vertices.setter
    def vertices(self, val):
        # assume when setting side lengths it must have been changed. no `If` which is slow.
        self._vertices = val
        self.area = poly_area(val)

    def _calc_ft(self) -> List[complex]:
        """
        Get fourier coefficients at given k points
        """
        return ft_2d_poly(self.vertices, self.ks)




# -*- coding: utf-8 -*-

from typing import Tuple, Optional
import time
import numpy as np
from inkstone.layer import Layer
from inkstone.sm import s_1l, s_1l_1212, s_1l_1221
from inkstone.layer_copy import layer_copy


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    from inkstone.params import Params
    from inkstone.mtr import Mtr

    lv = ((3., 0), (2., 4))
    num_g = 300
    omega = 0.5
    theta = 0.
    phi = 0.
    pr = Params(latt_vec=lv, num_g=num_g, omega=omega, theta=theta, phi=phi)

    vac = Mtr(epsi=1., mu=1., name='vacuum')
    di = Mtr(epsi=12., mu=1., name='di')
    materials = {'di': di, 'vacuum': vac}
    t = 0.4

    layer = Layer('my layer', t, 'vacuum', materials, pr)

    layer_copy = LayerCopy('copy', layer, 0.)

    layer.thickness = 0.6
    print(layer_copy.thickness)




# -*- coding: utf-8 -*-

import numpy as np
from inkstone.ft.ft_1d_sq import ft_1d_sq


def ft_2d_para(a, b, ks, center=(0, 0), shear_angle=90., rotate_angle=0.):
    """
    Calculate the fourier transform of a function that is 1 in side a parallelogram and 0 outside.

    The parallelogram has actual side lengths of a and b (b is on the ccw side of a).

    The parallelogram is rotated by `angle` and then shifted to `center`.

    Parameters
    ----------
    a               :   float
    b               :   float
                        side lengths of the parallelogram
    ks              :   list[tuple[float, float]]
    center          :   tuple[float, float]
    shear_angle     :   float
                        angle between a and b (a ccw rotates this angle to b)
    rotate_angle    :   float
                        ccw rotation angle of the whole shape in degrees

    Returns
    -------
    s               :   list[complex]
                        1d array
    """

    ia = np.pi * shear_angle / 180.
    b1 = np.sin(ia) * b
    m = np.tan(np.pi/2 - ia)

    shear = np.array([[1, m],
                      [0, 1]])

    ksa = np.array(ks)  # Nx2 array

    cen = np.array(center)

    ang = np.pi * rotate_angle / 180.

    rot = np.array([[np.cos(ang), -np.sin(ang)], [np.sin(ang), np.cos(ang)]])

    A = rot @ shear

    aksa = (A.T @ ksa.T).T

    sx = ft_1d_sq(a, aksa[:, 0])
    sy = ft_1d_sq(b1, aksa[:, 1])

    s = np.exp(-1j * cen @ ksa.T) * sx * sy

    return s.tolist()




# -*- coding: utf-8 -*-

import numpy as np
# import numpy.linalg as la
from inkstone.shps import Rect, Para, Disk, Elli, Poly, OneD
from warnings import warn
from typing import Tuple, Union, List, Optional
from inkstone.mtr import Mtr


class Bx:
    # Bx doesn't know what lattice it is in

    def __init__(self, mtr, shp, name=None, outside=None, **kwargs):
        """
        A box.
        The box has a name, its `mtr`, its `shp`.

        Can `set_shape`, set `ks`, and get Fourier series of epsi and mu by `fs()`.

        Parameters
        ----------
        mtr         :   Mtr
                        material
        shp         :   {'rectangle', 'disk', 'ellipse', 'polygon', 'parallelogram', '1d'}
        name        :   str
        outside     :   Bx
                        which bx contains this one.
        kwargs      :
                        other arguments for setting up the shape, to be passed on to `Rect`, `Disk`, `Elli`, `Poly`, or `OneD`.
        """

        self.epsi_ft = None
        self.epsi_inv_ft = None
        self.mu_ft = None
        self.mu_inv_ft = None

        self.mtr = mtr
        self.name = name

        if shp == 'rectangle':
            self.shp = Rect(**kwargs)
        elif shp == 'parallelogram':
            self.shp = Para(**kwargs)
        elif shp == 'disk':
            self.shp = Disk(**kwargs)
        elif shp == 'polygon':
            self.shp = Poly(**kwargs)
        elif shp == 'ellipse':
            self.shp = Elli(**kwargs)
        elif shp == '1d':
            self.shp = OneD(**kwargs)
        else:
            warn('shape not recognized.', UserWarning)

        self.outside: Bx = outside

    @property
    def shape(self):
        return self.shp.shape

    @property
    def ks(self):
        """Return the list of k points on which Fourier transform coefficients are calculated"""
        return self.shp.ks

    @ks.setter
    def ks(self, val):
        # After you set the ks here, Bx doesn't know if outside ks has changed. If it has, then you need to do ft(ks) to give the new ks.
        self.shp.ks = val

    def ft(self, ks=None, **kw_gibbs) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """
        Return the Fourier transform of the shape. The FT of a shape is understood as a function that is 1 inside and 0 outside.
        You can choose whether to have Gibbs correction for the returned FT coefficients.

        Parameters
        ----------
        ks          :   list
        kw_gibbs    :   keyword arguments related to Gibbs correction, to be passed on to `_calc_ft`

        Returns
        -------

        """
        if ks:
            self.ks = ks
        ep, ei, mu, mi = self._calc_ft(**kw_gibbs)
        self.epsi_ft = ep
        self.epsi_inv_ft = ei
        self.mu_ft = mu
        self.mu_inv_ft = mi

        return self.epsi_ft, self.epsi_inv_ft, self.mu_ft, self.mu_inv_ft

    def _calc_ft(self, **kw_gibbs) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """
        Calculate the Fourier transform of the box.

        Parameters
        ----------
        kw_gibbs    :   keyword arguments related to Gibbs correction, to be passed on to `.shps.Shp.ft`

        Returns
        -------

        """
        # t0 = time.process_time()
        _ft = np.array(self.shp.ft(**kw_gibbs), dtype=complex)
        # print(time.process_time()-t0)

        # no way of knowing if its material or that of the outer box changed
        # hence just recalculate every time. It takes no time anyway.
        epsi = self.mtr.epsi
        epsi_inv = self.mtr.epsi_inv
        mu = self.mtr.mu
        mu_inv = self.mtr.mu_inv
        if self.outside:
            epsi_out = self.outside.mtr.epsi
            epsi_out_inv = self.outside.mtr.epsi_inv
            mu_out = self.outside.mtr.mu
            mu_out_inv = self.outside.mtr.mu_inv
        else:
            epsi_out = np.eye(3, dtype=complex)
            epsi_out_inv = np.eye(3, dtype=complex)
            mu_out = np.eye(3, dtype=complex)
            mu_out_inv = np.eye(3, dtype=complex)
        epsi = epsi - epsi_out  # shouldn't do epsi -= epsi_out, which would be in-place change, which will change epsi of the material.
        epsi_inv = epsi_inv - epsi_out_inv
        mu = mu_inv - mu_out
        mu_inv = mu_inv - mu_out_inv

        # t1 = time.process_time()
        ep = epsi[None, :, :] * _ft[:, None, None]  # (N, 3, 3) shape
        ei = epsi_inv[None, :, :] * _ft[:, None, None]
        mu = mu[None, :, :] * _ft[:, None, None]
        mi = mu_inv[None, :, :] * _ft[:, None, None]
        # t2 = time.process_time()
        # print('ndarray time', t2-t1)
        # t1 = time.process_time()
        ep, ei, mu, mi = [[a for a in em] for em in [ep, ei, mu, mi]]
        # print('convert to list time', time.process_time()-t1)

        return ep, ei, mu, mi

    def set_shape(self, width=None, center=None, angle=None, shear_angle=None, half_lengths=None, side_lengths=None, radius=None, vertices=None, **kw_gibbs):
        """
        Set the shape geometry.

        Parameters
        ----------
        width           :   float
        center          :   tuple[float, float]
        angle           :   float
        shear_angle     :   float
        half_lengths    :   tuple[float, float]
        side_lengths    :   tuple[float, float]
        radius          :   float
        vertices        :   list[tuple[float, float]]
        kw_gibbs        :   other parameters for setting gibbs correction

        """
        # todo: what self.shp is is conditional; when refactoring this is a problem.
        if self.shp.shape == 'rectangle':
            if center is not None:
                self.shp.center = center
            if angle is not None:
                self.shp.angle = angle
            if side_lengths is not None:
                self.shp.side_lengths = side_lengths
        if self.shp.shape == 'parallelogram':
            if center is not None:
                self.shp.center = center
            if angle is not None:
                self.shp.angle = angle
            if shear_angle is not None:
                self.shp.shear_angle = shear_angle
            if side_lengths is not None:
                self.shp.side_lengths = side_lengths
        if self.shp.shape == 'disk':
            if center is not None:
                self.shp.center = center
            if radius is not None:
                self.shp.radius = radius
        if self.shp.shape == 'ellipse':
            if center is not None:
                self.shp.center = center
            if angle is not None:
                self.shp.angle = angle
            if half_lengths is not None:
                self.shp.half_lengths = half_lengths
        if self.shp.shape == 'polygon':
            if vertices is not None:
                self.shp.vertices = vertices
        if self.shp.shape == '1d':
            if width is not None:
                self.shp.width = width
            if center is not None:
                self.shp.center = center

        self.shp.use_gibbs_correction(**kw_gibbs)





# -*- coding: utf-8 -*-

import numpy as np


def ft_1d_sq(width, ks, center=0.):

    """
    Calculate the Fourier transform of a 1d square function.

    The square function looks like this:

           __________
    ______|    1     |____0____

    Parameters
    ----------
    center      :   float
                    center of the square
    width       :   float
                    width of the square
    ks          :   list
                    list of k points

    Returns
    -------
    s           :   list[complex]
                    1D array, Fourier series coefficients at the corresponding points
    """

    ksa = np.array(ks)

    s = np.exp(-1j * center * ksa) * width * np.sinc(ksa * width / 2. / np.pi)
    # note numpy sinc(x) definition is sin(pi x) / (pi x)

    return s.tolist()




# -*- coding: utf-8 -*-

from typing import List, Tuple, Union
import numpy as np


def pt_in_poly(vts: List[Tuple[float, float]],
               pt: Union[np.ndarray, Tuple[float, float]]) -> bool:
    """
    Determine if a point is in a polygon

    Parameters
    ----------
    vts     :   vertices of the polygon.
                The first point is also the last, no need to repeat it at the end of the list
    pt      :   coordinate of the point

    Returns
    -------
    yn      :   whether it is inside (True) or not (False)
    """
    yn = False
    x, y = pt
    for i, j in zip(vts, (vts[1:] + [vts[0]])):
        if (i[1] > y) != (j[1] > y) and (x < (i[0] + (j[0]-i[0]) * (y - i[1]) / (j[1] - i[1]))):
            yn = not yn

    return yn





# -*- coding: utf-8 -*-

import scipy.linalg as sla


def im(phi1, psi1, phi2, psi2, phi1_is_idt=False, psi1_is_idt=False):
    """
    calculate the interface matrix from layer 1 to layer 2
    This works for 2D TE, TM or 3D.

    Parameters
    ----------
    phi1        :   ndarray
    psi1        :   ndarray
    phi2        :   ndarray
    psi2        :   ndarray
    phi1_is_idt :   bool
                    if phi1 is identity matrix
    psi1_is_idt :   bool
                    if psi1 is identity matrix

    Returns
    -------
    a12     :   ndarray
    b12     :   ndarray

    """

    if phi1_is_idt:
        term1 = phi2
    else:
        term1 = sla.solve(phi1, phi2)
    if psi1_is_idt:
        term2 = psi2
    else:
        term2 = sla.solve(psi1, psi2)

    a12 = term1 + term2
    b12 = term1 - term2

    return a12, b12



# -*- coding: utf-8 -*-

import numpy as np
import numpy.linalg as la
from scipy import sparse as sps
import scipy.linalg as sla
# import scipy.fft as fft
from typing import Tuple, List, Union, Optional, Set
# import time
from warnings import warn
from inkstone.recipro import recipro
from inkstone.g_pts import g_pts
from inkstone.g_pts_1d import g_pts_1d
from inkstone.max_idx_diff import max_idx_diff
from inkstone.conv_mtx_idx import conv_mtx_idx_2d


class Params:
    """
    Contains various parameters that are relevant to solver and shared between solver and each layer.
    """

    def __init__(self,
                 latt_vec=None,
                 num_g=None,
                 frequency=None,
                 omega=None,
                 theta=None,
                 phi=None,
                 show_calc_time = False
                 ):
        """

        Parameters
        ----------
        latt_vec        :   tuple[tuple[float, float], tuple[float, float]]
                            primitive lattice vectors
        num_g           :   int
                            number of g points
        frequency       :   float or complex
        omega           :   float or complex
        theta, phi      :   float
                            incident angles in degrees.
                            Theta is the angle between incident k and z axis.
                            phi is the angle between the in-plane projection of k and the x axis. Rotating from kx axis phi degrees ccw around z axis to arrive at the kx of incident wave.
        """

        self.gs: Optional[List[Tuple[float, float]]] = None  # list of g points for E and H fields. Added by k_inci to get the ks, i.e. k points for E and H
        self.idx_g: Optional[List[Tuple[int, int]]] = None  # list of g points indices
        self.idx_conv_mtx: Optional[np.ndarray] = None  # indexing array to for constructing convolution matrices.
        self.ks: Optional[List[Tuple[float, float]]] = None  # list of k points for E and H fields.
        self.idx_g_ep_mu: Optional[List[Tuple[int, int]]] = None
        self.idxa_g_ep_mu: Optional[Tuple[np.ndarray, np.ndarray]] = None  # The g indices for ep and mu.
        self.idx_g_ep_mu_used: Optional[List[Tuple[int, int]]] = None  # the actually used epsilon and mu grid points
        self.ks_ep_mu: Optional[List[Tuple[float, float]]] = None  # list of k points where epsi and mu Fourier components are to be calculated
        self.ka_ep_mu: Optional[Tuple[np.ndarray, np.ndarray]] = None  # The k points for ep and mu, Tuple of meshgrid (kx, ky).
        self.Kx: Optional[np.ndarray] = None  # the diagonals of the Big diagonal matrix
        self.Ky: Optional[np.ndarray] = None  # the diagonals of the Big diagonal matrix
        self.Kz: Optional[np.ndarray] = None  # the diagonals of the Big diagonal matrix, useful for uniform layers
        self.mmax: Optional[int] = None  # max index in b1 direction for epsi and mu Fourier series
        self.nmax: Optional[int] = None  # max index in b2 direction for epsi and mu Fourier series
        self.phi0: Optional[np.ndarray] = None  # E field eigen mode in vacuum (identity matrix), side length 2*num_g
        self.psi0: Optional[np.ndarray] = None  # H field eigen mode in vacuum (Q * Phi * q^-1), side length 2*num_g
        self.phi0_2x2s: Optional[np.ndarray] = None  # the non-zero elements of phi0 stored in (2, 2, num_g) shape
        self.phif: Optional[np.ndarray] = None  # E field eigen mode in fic material, shape (2num_g, 2num_g)
        self.psif: Optional[np.ndarray] = None  # H field eigen mode in fic material, shape (2num_g, 2num_g)
        self.q0: Optional[np.ndarray] = None  # 1d array, eigen propagation constant in z direction in vacuum, length 2*num_g
        self.q0_half: Optional[np.ndarray] = None  # 1d array, eigen propagation constant in z direction in vacuum, length num_g
        self.q0_0: Optional[np.ndarray] = None  # 1d array, containing idxs to the idx_g list which q0 is 0, i.e. parallel to surface
        self.q0_inv: Optional[np.ndarray] = None  # 1d array, 1./q0, elementwise inversion of q0, length 2*num_g
        self.P0_val: Optional[Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]] = None  # Tuple of 4, each is an ndarray of size num_g, containing the diagonal elements of the 4 blocks of P0.
        self.Q0_val: Optional[Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]] = None  # Tuple of 4, each is an ndarray of size num_g, containing the diagonal elements of the 4 blocks of Q0.
        self.im0: Optional[Tuple[np.ndarray, np.ndarray]] = None  # vacuum interface matrix
        self.sm0: Optional[Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]] = None  # vacuum scattering matrix, (s11_0, s12_0, s21_0, s22_0), each of s_ij has side length 2*num_g
        self._rad_cha_0: Optional[List[int]] = None

        self.ccnif = "physical"
        self.ccpif = "ac"

        self.__num_g_actual: Optional[int] = None  # actual number of G points used.
        self._num_g_input: Optional[int] = None  # number of G points input by user
        self._k_inci: Optional[Tuple[Tuple[float, float]], Tuple[float, float]] = None  # incident in-plane k
        self._omega: Optional[complex] = None
        self._frequency: Optional[complex] = None
        self._recipr_vec: Optional[Tuple[Tuple[float, float], Tuple[float, float]]] = None
        self._latt_vec: Optional[Tuple[Tuple[float, float], Tuple[float, float]]] = None
        self.is_1d_latt: bool = False  # if this is solving a 2d structure. Incident wave can still have phi rotation, i.e. 3D scenario for a 2D structure.
        self._theta: Optional[float] = None  # in rad
        self._phi: Optional[float] = None  # in rad
        self._uc_area: Optional[float] = None
        self.cos_varthetas: Optional[List[float]] = None  # incident angles
        self.sin_varthetas: Optional[List[float, complex]] = None
        self.cos_phis: Optional[List[float]] = None
        self.sin_phis: Optional[List[float]] = None
        self.cos_varthetas_bk: Optional[List[float]] = None  # incident angles
        self.sin_varthetas_bk: Optional[List[float, complex]] = None
        self.cos_phis_bk: Optional[List[float]] = None
        self.sin_phis_bk: Optional[List[float]] = None

        self._incident_orders = None
        self._incident_orders_bk = None
        self._s_amps = None
        self._p_amps = None
        self._s_amps_bk = None
        self._p_amps_bk = None
        self.ai: Optional[np.ndarray] = None
        self.bo: Optional[np.ndarray] = None

        # self.q0_contain_0: bool = False

        self.show_calc_time = show_calc_time

        self._inci_is_iso_nonvac = False
        self._ind_inci = 1.
        self._out_is_iso_nonvac = False
        self._ind_out = 1.

        # property initialization can change other attributes hence must be after initialization of other attributes.
        if latt_vec is not None:
            self.latt_vec = latt_vec
        if num_g is not None:
            self.num_g = num_g
        if frequency is not None:
            self.frequency = frequency
        if omega is not None:
            self.omega = omega
        if theta is not None:
            self.theta = theta
        if phi is not None:
            self.phi = phi

    @property
    def latt_vec(self) -> Union[float, Tuple[Tuple[float, float], Tuple[float, float]]]:
        """
        Primitive lattice vectors.
        The lattice vector and be a single number which means 1D, or tuples such as ((a1x, a1y), (a2x, a2y)) which means 2D .

        If tuples are supplied but one of them is (0, 0), then it is also a 1d. In this case the lattice will be rotated such that the actual 1d direction is in x axis where y is uniform.
        Note the incident wave angles will be w.r.t. the structure AFTER it's rotated to x axis.

        """
        return self._latt_vec

    @latt_vec.setter
    def latt_vec(self, val: Union[float, Tuple[Tuple[float, float], Tuple[float, float]]]):
        if type(val) is not tuple:
            val = ((val, 0), (0, 0))
            self.is_1d_latt = True
        else:
            a, b = val
            an, bn = [la.norm(a), la.norm(b)]
            # if either latt vec is zero, then it is 1D. Make sure it's x-z for layer solver.
            if an == 0:
                val = ((bn, 0), (0, 0))
                self.is_1d_latt = True
                warn("2D structure (in-plane 1D), non-uniform direction is rotated to x axis.")
            if bn == 0:
                val_old = val
                val = ((an, 0), (0, 0))
                self.is_1d_latt = True
                if val_old != val:
                    warn("2D structure (in-plane 1D), non-uniform direction is rotated to x axis.")
        self._latt_vec: Union[Tuple[Tuple[float, float], Tuple[float, float]]] = val
        self._recipr_vec: Tuple[Tuple[float, float], Tuple[float, float]] = recipro(val[0], val[1])
        self.if_2d()
        self._calc_gs()
        self._calc_uc_area()
        # self._calc_ks_ep_mu()  # called through _calc_gs -> _calc_ks

    @property
    def recipr_vec(self) -> Tuple[Tuple[float, float], Tuple[float, float]]:
        """Reciprocal lattice vectors"""
        return self._recipr_vec

    def if_2d(self):
        b1, b2 = self._recipr_vec
        if b1[0] == float('inf') or b1[1] == float('inf') or b2[0] == float('inf') or b2[1] == float('inf'):
            self.is_1d_latt = True
        else:
            self.is_1d_latt = False

    @property
    def num_g(self) -> int:
        """The number of G points to be included. To ensure symmetry, the actual num_g may be slightly different than the user set value."""
        return self._num_g_ac

    @num_g.setter
    def num_g(self, val: int):
        self._num_g_input: int = val
        self._calc_gs()

    @property
    def _num_g_ac(self) -> int:
        """actual number of G points"""
        return self.__num_g_actual

    @_num_g_ac.setter
    def _num_g_ac(self, val: int):
        # todo: all other calc should use this than original num_g
        self.__num_g_actual = val
        # self._calc_phi0()
        self._calc_phi0_psi0()
        self._calc_phif_psif()
        # self._calc_q0()  # called through _calc_gs - _calc_ks
        self._calc_im0()
        self._calc_s_0()
        # self._calc_ai_bo_3d()  # called through _calc_gs

    @property
    def frequency(self) -> Union[float, complex]:
        """ frequency, c/lambda in vacuum """
        return self._frequency

    @frequency.setter
    def frequency(self, val: Union[float, complex]):
        if val is not None:
            self.omega = val * np.pi * 2.
            self._frequency = val

    @property
    def omega(self) -> Union[float, complex]:
        """ angular frequency, 2pi c/lambda in vacuum """
        return self._omega

    @omega.setter
    def omega(self, val: Union[float, complex]):
        if val is not None:
            self._omega = val
            self._frequency = val / np.pi / 2.
            # self.q0_contain_0 = False
            # self._calc_gs()  # recalculating gs because some g may be possibly removed in previous runs to remove Wood's anomaly.
            self._calc_k_inci()
            # self._calc_q0()  # called through _calc_k_inci - _calc_ks
            # self._calc_P0Q0()  # called through _calc_k_inci - _calc_ks - _calc_Km
            # self._calc_angles()  # called through _calc_k_inci - _calc_ks -

    @property
    def k_inci(self) -> Tuple[float, float]:
        """kx and ky projection of the incident wave vector"""
        return self._k_inci

    @k_inci.setter
    def k_inci(self, val: Tuple[float]):
        """Directly setting incident in-plane kx and ky"""
        self._k_inci = val
        self._calc_ks()

    @property
    def theta(self) -> Union[float, complex]:
        """he angle between incident k and z axis, in degrees, range: [0, pi/2]"""
        if self._theta is not None:
            theta = self._theta / np.pi * 180.
        else:
            theta = None
        return theta

    @theta.setter
    def theta(self, val: Union[float, complex]):
        if val is not None:
            self._theta: float = val * np.pi / 180.
            self._calc_k_inci()
            self._calc_angles()
        else:
            if self._theta is None:
                self._theta = 0.

    @property
    def phi(self) -> float:
        """ the angle between the in-plane projection of k and the x axis in degrees. Rotating from kx axis phi degrees ccw around z axis to arrive at the kx of incident wave. """
        if self._phi is not None:
            phi = self._phi / np.pi * 180.
        else:
            phi = None
        return phi

    @phi.setter
    def phi(self, val: float):
        if val is not None:
            self._phi: float = val * np.pi / 180.
            self._calc_k_inci()
            self._calc_angles()
        else:
            if self._phi is None:
                self._phi = 0.

    @property
    def rad_cha_0(self) -> List[int]:
        """list of indices in `self.idx_g` that corresponds to radiation channels in vacuum"""
        return self._rad_cha_0  # between 0 and num_g

    @property
    def ccnif(self) -> str:
        """
        {"physical", "ac"}, default: "physical".
        """
        return self._ccnif

    @ccnif.setter
    def ccnif(self, val):
        if (val == "physical") or (val == "ac"):
            self._ccnif = val
        else:
            warn("ccnif not understood and not changed in the solver.")

    @property
    def ccpif(self) -> str:
        """
        {"physical", "ac"}, default: "ac".
        """
        return self._ccpif

    @ccpif.setter
    def ccpif(self, val):
        if (val == "physical") or (val == "ac"):
            self._ccpif = val
        else:
            warn("ccpif not understood and not changed in the solver.")

    def set_inci_ord_amp(self,
                         s_amplitude: Optional[Union[float, complex, List[Union[float, complex]]]] = None,
                         p_amplitude: Optional[Union[float, complex, List[Union[float, complex]]]] = None,
                         order: Union[Tuple[int, int], List[Tuple[int, int]]] = None,
                         s_amplitude_back: Union[float, List[float]] = None,
                         p_amplitude_back: Union[float, List[float]] = None,
                         order_back: Union[Tuple[int, int], List[Tuple[int, int]]] = None
                         ):
        """
        Setting and resetting incident orders and amplitudes.

        If any of the optional arguments are not given, then use the previously set values.

        Parameters
        ----------
        s_amplitude
        p_amplitude
        order
        s_amplitude_back
        p_amplitude_back
        order_back
        """
        # Convert order and order_back to lists (can be empty list)
        # todo: how do I know the user chosen order is in idx_g?
        if order is None:
            if self._incident_orders is None:
                self._incident_orders = [(0, 0)]
        elif type(order) is tuple:
            order = [order]
            self._incident_orders = order
        else:
            self._incident_orders = order

        if order_back is None:
            if self._incident_orders_bk is None:
                self._incident_orders_bk = [(0, 0)]
        elif type(order_back) is tuple:
            order_back = [order_back]
            self._incident_orders_bk = order_back
        else:
            self._incident_orders_bk = order_back

        # convert s and p amplitudes to lists (can be empty list)
        amp = []
        for amps, selfamps, od in zip([[s_amplitude, p_amplitude], [s_amplitude_back, p_amplitude_back]], [[self._s_amps, self._p_amps], [self._s_amps_bk, self._p_amps_bk]], [self._incident_orders, self._incident_orders_bk]):
            for a, s in zip(amps, selfamps):
                if a is None:
                    if s is None:
                        a = [0. for i in range(len(od))]
                    else:
                        a = s
                elif not hasattr(a, "__len__"):
                    a = [a]
                amp.append(a)
        self._s_amps = amp[0]
        self._p_amps = amp[1]
        self._s_amps_bk = amp[2]
        self._p_amps_bk = amp[3]

        if (len(self._incident_orders) != len(self._s_amps)) or (len(self._incident_orders) != len(self._p_amps)):
            raise Exception("The list length of the incident s amplitudes, p amplitudes, and incident orders are not equal.")

        if (len(self._incident_orders_bk) != len(self._s_amps_bk)) or (len(self._incident_orders_bk) != len(self._p_amps_bk)):
            raise Exception("The list length of the backside incident s amplitudes, p amplitudes, and incident orders are not equal.")

        # self._calc_ai_bo_3d()

    @property
    def inci_is_iso_nonvac(self) -> bool:
        """if inci region is isotropic non-vacuum"""
        return self._inci_is_iso_nonvac

    @inci_is_iso_nonvac.setter
    def inci_is_iso_nonvac(self, val):
        self._inci_is_iso_nonvac = val
        self._calc_k_inci()

    @property
    def ind_inci(self) -> float:
        """inci region refractive index"""
        return self._ind_inci

    @ind_inci.setter
    def ind_inci(self, val):
        self._ind_inci = val
        self._calc_k_inci()

    @property
    def out_is_iso_nonvac(self) -> bool:
        """if out region is isotropic non-vacuum"""
        return self._out_is_iso_nonvac

    @out_is_iso_nonvac.setter
    def out_is_iso_nonvac(self, val):
        self._out_is_iso_nonvac = val
        self._calc_k_inci()

    @property
    def ind_out(self) -> float:
        """out region refractive index"""
        return self._ind_out

    @ind_out.setter
    def ind_out(self, val):
        self._ind_out = val
        self._calc_k_inci()

    def _calc_k_inci(self):
        """calculate incident kx and ky"""
        if (self.theta is not None) and (self.phi is not None) and (self.omega is not None):
            if self.inci_is_iso_nonvac:
                k = self.omega.real * self.ind_inci  # todo: what if ind_inci complex?
            else:
                k = self.omega.real
            # todo: for complex freq, should here be np.abs(self.omega) or self.omega.real?
            # todo: using self.omega (hence kx ky complex) also gives answers, but the physical meaning is different
            kx = k * np.cos(np.pi/2 - self._theta) * np.cos(self._phi)
            ky = k * np.cos(np.pi/2 - self._theta) * np.sin(self._phi)
            self._k_inci: Tuple[float, float] = (kx, ky)
            self.k_inci = (kx, ky)

            # self._calc_ks()  # called in `k_inci` setter

    def _calc_gs(self):
        """ calculate E and H Fourier components g points """
        if self._num_g_input and self.recipr_vec:
            b1, b2 = self.recipr_vec
            b1n, b2n = [la.norm(b) for b in [b1, b2]]
            if b1n != float('inf') and b2n != float('inf'):
                self.gs, self.idx_g = g_pts(self._num_g_input, self.recipr_vec[0], self.recipr_vec[1])
            elif b1n == float('inf'):
                self.gs, idx = g_pts_1d(self._num_g_input, b2)
                self.idx_g = [(0, i) for i in idx]
            elif b2n == float('inf'):
                self.gs, idx = g_pts_1d(self._num_g_input, b1)
                self.idx_g = [(i, 0) for i in idx]
            else:
                raise Exception("Both reciprocal lattice vectors are infinite. Can't calculate g points.")
            self._num_g_ac = len(self.gs)
            self._calc_ks()
            self._calc_conv_mtx_idx()
            # self._calc_ai_bo_3d()  # called through _calc_ks() - _calc_angles()

    def _remove_gs(self,
                   idxs_rm: List[Tuple[int, int]]):
        """
        remove certain g points by given list of g indices such as [(1, -2)]

        Parameters
        ----------
        idxs_rm:
            indices to remove, each element in this list should be a tuple of integers like (1, -2)

        Returns
        -------

        """
        idx_g_new = []
        gs_new = []
        for ii, idx in enumerate(self.idx_g):
            to_add = True
            for idx_rm in idxs_rm:
                if idx == idx_rm:
                    to_add = False
                    break
            if to_add:
                idx_g_new.append(idx)
                gs_new.append(self.gs[ii])
        self.idx_g = idx_g_new
        self.gs = gs_new
        self._num_g_ac = len(self.gs)
        self._calc_ks()
        self._calc_conv_mtx_idx()

    def _remove_gs_xuhao(self,
                         idxs: List[int]):
        """
        Remove g points by the index in the current idx_g list, for example, idxs=[3, 5] means to remove the 3rd and the 5th element in the g list.

        Parameters
        ----------
        idxs

        Returns
        -------

        """
        for ii in sorted(idxs, reverse=True):
            del self.gs[ii]
            del self.idx_g[ii]
        self._num_g_ac = len(self.gs)
        self._calc_ks()
        self._calc_conv_mtx_idx()

    def _calc_ks(self):
        if self.gs and self.k_inci:
            self.ks = [(g[0]+self.k_inci[0], g[1] + self.k_inci[1]) for g in self.gs]
            self._calc_q0()
            self._calc_Km()
            self._calc_ks_ep_mu()
            self._calc_angles()

    def _calc_Km(self):
        """Calculate Kx Ky arrays"""
        if self.ks:
            # t1 = time.process_time()
            ksa = np.array(self.ks)  # shape (NumG, 2)
            kx = ksa[:, 0]
            ky = ksa[:, 1]
            self.Kx = kx
            self.Ky = ky

            # print('_calc_Km', time.process_time() - t1)

            self._calc_P0Q0()
            # self._calc_phi0_psi0()  # called through _calc_P0Q0()

    def _calc_conv_mtx_idx(self):
        if self.idx_g:
            self.idx_conv_mtx = conv_mtx_idx_2d(self.idx_g, self.idx_g)
            self.idx_g_ep_mu_used = list(set([(i, j) for (i, j) in self.idx_conv_mtx.reshape(self._num_g_ac ** 2, 2)]))  # The first element of each tuple is in physical "x" direction

    def _calc_ks_ep_mu(self):
        if self.idx_g:
            # t1 = time.process_time()

            m, n = max_idx_diff(self.idx_g)
            self.mmax = m
            self.nmax = n
            x = np.arange(-m, m+1)
            y = np.arange(-n, n+1)
            xx, yy = np.meshgrid(x, y)

            # t2 = time.process_time()

            self.idx_g_ep_mu = list(zip(xx.ravel(), yy.ravel()))
            self.idxa_g_ep_mu = (xx, yy)

            # t3 = time.process_time()

            b1, b2 = self.recipr_vec
            b1n, b2n = [la.norm(b) for b in [b1, b2]]
            # if one of them is infinity, this is a 1D structure. set it to 0 such that they don't appear in ks_ep_mu
            if b1n == float('inf'):
                b1 = (0, 0)
            if b2n == float('inf'):
                b2 = (0, 0)

            kx = xx * b1[0] + yy * b2[0]
            ky = xx * b1[1] + yy * b2[1]
            self.ks_ep_mu = list(zip(kx.ravel(), ky.ravel()))  # each element is (kx, ky) tuple.
            self.ka_ep_mu = (kx, ky)

            # print(t2 - t1)
            # print(t3 - t2)
            # print(time.process_time() - t3)
            # print('_calc_ks_ep_mu', time.process_time()-t1)

    def _calc_q0(self):
        if self._num_g_ac and (self.omega is not None) and self.ks:
            # t1 = time.process_time()
            k_parallel = np.linalg.norm(np.array(self.ks), axis=-1)
            q02 = np.ones(self._num_g_ac) * np.square(self.omega) - np.square(k_parallel) + 0j
            self._rad_cha_0 = np.where(q02.real > 0)[0].tolist()  # todo: even for radiation channel, if omega.imag larger than omega.real, q02.real is negative
            q0 = np.sqrt(q02)
            if self.omega.imag < 0:
                if self.ccnif == "physical":
                    q0[(q02.real < 0) * (q0.imag < 0)] *= -1
                elif self.ccnif == "ac":
                    q0[(q02.real < 0) * (q0.imag > 0)] *= -1
                else:
                    warn("ccnif not recognized. Default to 'physical'.")
                    q0[(q02.real < 0) * (q0.imag < 0)] *= -1
            elif self.omega.imag > 0:
                if self.ccpif == "ac":
                    q0[(q02.real < 0) * (q0.imag < 0)] *= -1
                elif self.ccpif == "physical":
                    q0[(q02.real < 0) * (q0.imag > 0)] *= -1
                else:
                    warn("ccpif not recognized. Default to 'ac'.")
                    q0[(q02.real < 0) * (q0.imag < 0)] *= -1
            else:
                q0[q0.imag < 0] *= -1
            # todo: what to do at q02.real == 0 case (Woods)?

            self.q0_0 = np.where(np.abs(q0) == 0.)[0]

            self.q0 = np.concatenate([q0, q0])
            self.q0_half = q0
            # print('_calc_q0', time.process_time() - t1)

            with np.errstate(divide='ignore', invalid='ignore'):
                self.q0_inv = 1. / self.q0
                ii = np.where(self.q0 == 0.)
                self.q0_inv[ii] = float('inf')

            # self._calc_psi0()
            self._calc_phi0_psi0()

    def _calc_P0Q0(self):
        if (self.omega is not None) and (self.Kx is not None) and (self.Ky is not None):
            # t1 = time.process_time()

            Kx = self.Kx
            Ky = self.Ky
            o = self.omega

            P11 = 1. / o * Kx * Ky
            P12 = o - 1. / o * Kx * Kx
            P21 = -o + 1. / o * Ky * Ky
            P22 = - 1. / o * Ky * Kx

            # t2 = time.process_time()
            # print('P0 Q0 blocks', t2 - t1)

            self.P0_val = (P11, P12, P21, P22)
            self.Q0_val = self.P0_val

            # # for debugging
            # ng = self._num_g_ac
            # self.P0 = np.zeros((2*ng, 2*ng), dtype=complex)
            # r1 = range(ng)
            # r2 = range(ng, 2 * ng)
            # self.P0[r1, r1] = P11
            # self.P0[r2, r1] = P21
            # self.P0[r1, r2] = P12
            # self.P0[r2, r2] = P22
            # self.P0 = np.block([[np.diag(P11), np.diag(P12)],
            #                     [np.diag(P21), np.diag(P22)]])

            # self._calc_psi0()
            self._calc_phi0_psi0()
            self._calc_phif_psif()

    def _calc_phi0_psi0(self):
        if (self.Q0_val is not None) and (self.q0 is not None) and (self._num_g_ac is not None):
            # t1 = time.process_time()

            # o = self.omega
            # Kx = self.Kx.copy()
            # Ky = self.Ky.copy()
            # k_norm = np.sqrt(np.conj(Kx) * Kx + np.conj(Ky) * Ky)

            # # Use [ky, -kx] etc. as eigen
            # Tx = Kx  # term with x subscript
            # Ty = Ky
            #
            # # for small k just use [1, 0], [0, 1] as eigen.
            # skc = 0.5  # small k criterion
            # small_k = np.where(k_norm < (skc * np.abs(o)))[0]
            # Tx[small_k] = 0.
            # Ty[small_k] = 1.
            # t_norm = k_norm.copy()  # term norm
            # t_norm[small_k] = 1.
            #
            # Txm = np.diag(Tx)
            # Tym = np.diag(Ty)
            # jwTxq = 1j / o * Tx * self.q0_half
            # jwTyq = 1j / o * Ty * self.q0_half
            #
            # # for small k, need separate calc here
            # jwTxq[small_k] = -1j / o * Kx[small_k] * Ky[small_k] / self.q0_half[small_k]
            # jwTyq[small_k] = -1j / o * (-np.conj(Kx[small_k])*Kx[small_k] - np.conj(self.q0_half[small_k])*self.q0_half[small_k]) / self.q0_half[small_k]
            #
            # jwTxqm = np.diag(jwTxq)
            # jwTyqm = np.diag(jwTyq)
            #
            # ng = self._num_g_ac
            # r1 = range(ng)
            # r2 = range(ng, 2 * ng)
            # phi0 = np.zeros((2*ng, 2*ng), dtype=complex)
            # psi0 = phi0.copy()
            # phi0[r1, r1] = Ty
            # phi0[r2, r1] = -Tx
            # phi0[r1, r2] = jwTxq
            # phi0[r2, r2] = jwTyq
            # psi0[r1, r1] = jwTxq
            # psi0[r2, r1] = jwTyq
            # psi0[r1, r2] = Ty
            # psi0[r2, r2] = -Tx
            #
            # # phi0 = np.block([[Tym, jwTxqm],
            # #                  [-Txm, jwTyqm]])
            # # psi0 = np.block([[jwTxqm, Tym],
            # #                  [jwTyqm, -Txm]])
            #
            # self.phi0_2x2s = np.array([[Ty, jwTxq],
            #                           [-Tx, jwTyq]])


            # with normalization
            q0h = self.q0_half
            ng = self._num_g_ac
            o = self.omega
            Kx = self.Kx.copy()
            Ky = self.Ky.copy()
            k_norm = np.sqrt(np.conj(Kx) * Kx + np.conj(Ky) * Ky)

            # skc = 0.05
            # i_knz = np.where(k_norm < (skc * np.abs(o)))[0]
            i_kez = np.where(k_norm == 0.)[0]
            # i_qsw = np.where((np.abs(q0h) <= np.abs(o)) * (k_norm >= (skc * np.abs(o))))[0]
            i_qsw = np.where((np.abs(q0h) < np.abs(o)))[0]
            i_qlw = np.where(np.abs(q0h) > np.abs(o))[0]
            # when allowing imaginary/complex kx, ky: when ky = i kx, q = omega, but k_norm is not zero. But this case doesn't matter. as long as k_norm not zero, it can be divided.
            idxa = np.array(self.idx_g)
            ii = (idxa[:, 0] == 0) & (idxa[:, 1] == 0)

            c1 = np.array([Ky, -Kx], dtype=complex)
            c2 = np.array([Kx, Ky], dtype=complex)
            c1f = np.ones(ng, dtype=complex)
            c2f = c1f.copy()

            # c1[:, i_knz] = np.array([[1.], [0.]])
            # c2[:, i_knz] = -1j / o * np.array([Kx[i_knz] * Ky[i_knz] / q0h[i_knz], (-np.square(Kx[i_knz]) - np.square(q0h[i_knz])) / q0h[i_knz]])  # should not be |Kx|^2
            c1[:, i_kez] = np.array([[1.], [0.]], dtype=complex)
            c2[:, i_kez] = np.array([[0.], [1.]], dtype=complex)
            cphi = np.cos(self._phi)
            sphi = np.sin(self._phi)
            c1[:, ii] = np.array([[sphi], [-cphi]], dtype=complex)
            c2[:, ii] = np.array([[cphi], [sphi]], dtype=complex)

            c1f[i_qlw] = o / q0h[i_qlw] / k_norm[i_qlw]
            c2f[i_qlw] = 1j / k_norm[i_qlw]

            c1f[i_qsw] = 1. / k_norm[i_qsw]
            c2f[i_qsw] = 1j / o * q0h[i_qsw] / k_norm[i_qsw]

            # c1f[i_knz] = 1.
            # c2f[i_knz] = 1.

            c1f[i_kez] = 1.
            c2f[i_kez] = 1j * q0h[i_kez] / o

            c1f[ii] = 1.
            c2f[ii] = 1j * q0h[ii] / o

            c1 *= c1f
            c2 *= c2f

            r1 = range(ng)
            r2 = range(ng, 2 * ng)
            phi0 = np.zeros((2*ng, 2*ng), dtype=complex)
            psi0 = phi0.copy()
            phi0[r1, r1] = c1[0, :]
            phi0[r2, r1] = c1[1, :]
            phi0[r1, r2] = c2[0, :]
            phi0[r2, r2] = c2[1, :]
            psi0[r1, r1] = c2[0, :]
            psi0[r2, r1] = c2[1, :]
            psi0[r1, r2] = c1[0, :]
            psi0[r2, r2] = c1[1, :]

            self.phi0_2x2s = np.moveaxis(np.array([c1, c2]), 0, 1)

            # # debugging,  check if phi is eigen and consistent with psi
            # psi00 = -1j * self.P0 @ phi0 / self.q0
            # diff = np.abs(psi00 - psi0).max()
            # print('psi0 diff {:g}'.format(diff))
            #
            # check_eigen = self.P0 @ self.P0 @ phi0
            # diff1 = np.abs(check_eigen + phi0 * self.q0 * self.q0).max()
            # print('check eigen {:g}'.format(diff1))

            # # original
            # phi0 = np.eye(2 * self._num_g_ac, 2 * self._num_g_ac, dtype=complex)
            # ng = self._num_g_ac
            # psi0 = np.zeros((2 * ng, 2 * ng), dtype=complex)
            # r1 = range(ng)
            # r2 = range(ng, 2 * ng)
            # q0_inv = self.q0_inv
            # if self.Q0_val[0].size == self.q0_inv[:ng].size:
            #     psi0[r1, r1] = -1j * self.Q0_val[0] * q0_inv[:ng]
            #     psi0[r1, r2] = -1j * self.Q0_val[1] * q0_inv[ng:]
            #     psi0[r2, r1] = -1j * self.Q0_val[2] * q0_inv[:ng]
            #     psi0[r2, r2] = -1j * self.Q0_val[3] * q0_inv[ng:]

            self.phi0 = phi0
            self.psi0 = psi0

    def _calc_phif_psif(self):
        if self._num_g_ac:
            ng = self._num_g_ac

            phif = np.eye(2 * ng, 2 * ng, dtype=complex)
            psif = np.zeros((2 * ng, 2 * ng), dtype=complex)

            r1 = range(ng)
            r2 = range(ng, 2 * ng)

            # attention! phif is assumed to be identity in interface calculations. if need to change the form, you need to change coding there, not just changing phif here.
            # phif[r2, r2] = -1.

            # attention! If need to change this form, note in interface matrix calculations this form is assumed to speed up things. Need to change coding there, not just changing psif here.
            psif[r2, r1] = 1.j
            psif[r1, r2] = -1.j

            self.phif = phif
            self.psif = psif

            # # for debugging
            # phif2 = np.eye(2 * ng, 2 * ng, dtype=complex)
            # psif2 = np.zeros((2 * ng, 2 * ng), dtype=complex)
            # phif2[r2, r2] = -1.  # attention! this was for debugging, but note since phif=eye, it is omitted in certain calculations
            # psif2[r2, r1] = 1.
            # psif2[r1, r2] = -1.
            # self.phif2 = phif2
            # self.psif2 = psif2

            # phif3 = np.zeros((2 * ng, 2 * ng), dtype=complex)
            # psif3 = np.zeros((2 * ng, 2 * ng), dtype=complex)
            # phif3[r1, r1] = 1.
            # phif3[r2, r1] = 1j
            # phif3[r1, r2] = -1j
            # phif3[r2, r2] = 1.
            # psif3 = phif3
            #
            # self.phif3 = phif3
            # self.psif3 = psif3

            # if self.P0_val is not None:
            #     phif4 = np.random.rand(2*ng, 2*ng)
            #     psif4 = -1j * self.P0 @ phif4 * self.q0_inv
            #     self.phif4 = phif4
            #     self.psif4 = psif4

            # # end of debugging

    def _calc_phi0(self):
        warn('This method is deprecated. Use `_calc_phi0_psi0` instead.', category=DeprecationWarning)
        if self._num_g_ac:
            self.phi0 = np.eye(2 * self._num_g_ac, 2 * self._num_g_ac, dtype=complex)

    def _calc_psi0(self):
        warn('This method is deprecated. Use `_calc_phi0_psi0` instead.', category=DeprecationWarning)

        if (self.Q0_val is not None) and (self.q0 is not None) and (self._num_g_ac is not None) and ((self.Kx is not None) and (self.Ky is not None)):
            # t1 = time.process_time()

            # if not self.q0_contain_0:
            ng = self._num_g_ac
            psi0 = np.zeros((2 * ng, 2 * ng), dtype=complex)
            r1 = range(ng)
            r2 = range(ng, 2 * ng)
            q0_inv = self.q0_inv
            if self.Q0_val[0].size == self.q0_inv[:ng].size:
                psi0[r1, r1] = -1j * self.Q0_val[0] * q0_inv[:ng]
                psi0[r1, r2] = -1j * self.Q0_val[1] * q0_inv[ng:]
                psi0[r2, r1] = -1j * self.Q0_val[2] * q0_inv[:ng]
                psi0[r2, r2] = -1j * self.Q0_val[3] * q0_inv[ng:]

                # # at Wood P is singular so this solving doesn't work.
                # _P = [np.diag(Qv) for Qv in self.Q0_val]
                # P = np.block([[_P[0], _P[1]],
                #               [_P[2], _P[3]]])
                # psi0 = 1j * la.solve(P, self.phi0) @ np.diag(self.q0)

            o = self.omega
            cn = np.where(np.abs(self.q0) == 0.)[0]
            cn1 = np.where(np.abs(self.q0) < 1e-2)[0]  # for debugging

            # # test setting zeros an ones
            # for ii in cn:
            #     if ii < ng:
            #         if self.Ky[ii] == 0.:
            #             psi0[ii, ii] = 0.
            #             psi0[ii+ng, ii] = 1.
            #             self.phi0[ii, ii] = 0.
            #         elif self.Kx[ii] != 0.:
            #             # psi0[ii, ii] = -self.P0_val[0][ii]
            #             # psi0[ii+ng, ii] = -self.P0_val[2][ii]
            #             # self.phi0[ii, ii] = 0.
            #             if (ii == 5) or (ii == 6):
            #                 psi0[ii, ii] = 0.
            #                 psi0[ii+ng, ii] = 0.
            #                 self.phi0[ii, ii] = 1.
            #                 self.phi0[ii+ng, ii] = -1.
            #
            #         else:
            #             psi0[ii, ii] = 1. * np.sign(self.Ky[ii])
            #             psi0[ii+ng, ii] = 0.
            #
            #         # if (self.Ky[ii] == 0) or (self.Kx[ii] == 0):
            #         #     # psi0[ii, ii] = -1j * 1. / o * self.Kx[cn]
            #         #     psi0[ii, ii] = 0.
            #         #     psi0[ii+ng, ii] = 1.
            #         # else:
            #         #     psi0[ii, ii] = self.P0_val[0][ii]
            #         #     psi0[ii+ng, ii] = self.P0_val[2][ii]
            #     else:
            #         iim = ii-ng
            #         if self.Kx[iim] == 0.:
            #             psi0[ii, ii] = 0.
            #             psi0[ii-ng, ii] = -1.
            #             self.phi0[ii, ii] = 0.
            #         elif self.Ky[iim] != 0.:
            #             # psi0[ii-ng, ii] = -self.P0_val[2][iim]
            #             # psi0[ii, ii] = -self.P0_val[3][iim]
            #             # self.phi0[ii, ii] = 0.
            #             if (ii == 14) or (ii == 15):
            #                 psi0[ii, ii] = 1.
            #                 psi0[ii-ng, ii] = -1.
            #                 self.phi0[ii, ii] = 0.
            #                 self.phi0[ii-ng, ii] = -0.
            #
            #         else:
            #             psi0[ii, ii] = 1. * np.sign(self.Kx[iim])
            #             psi0[ii-ng, ii] = 0.
            #
            #         # if (self.Ky[iim] == 0) or (self.Kx[iim] == 0):
            #         #     psi0[ii, ii] = 0.
            #         #     psi0[ii-ng, ii] = 1.
            #         # else:
            #         #     psi0[ii-ng, ii] = self.P0_val[1][iim]
            #         #     psi0[ii, ii] = self.P0_val[3][iim]
            #
            #     # self.phi0[ii, ii] = 0.

            # # this part is to test scaling of Phi and Psi
            # factor = 1e-7
            # psi0[5, 5] *= factor
            # psi0[14, 5] *= factor
            # self.phi0[5, 5] *= factor
            # for ii in cn1:
            #     if ii < ng:
            #         if (self.Kx[ii] != 0.) and (self.Ky[ii] != 0.):
            #             psi0[ii, ii] *= factor
            #             psi0[ii + ng, ii] *= factor
            #             self.phi0[ii, ii] *= factor * 1e-0
            #     else:
            #         iim = ii-ng
            #         if (self.Kx[iim] != 0.) and (self.Ky[iim] != 0.):
            #             psi0[ii-ng, ii] *= factor
            #             psi0[ii, ii] *= factor
            #             self.phi0[ii, ii] *= factor * 1e-0
            #
            # for ii in cn1:
            #     if ii < ng:
            #         if (self.Kx[ii] != 0.) and (self.Ky[ii] != 0.):
            #             psi0[ii, ii] = -1j
            #             psi0[ii + ng, ii] = 1j
            #             self.phi0[ii, ii] = 0.
            #             self.phi0[ii+ng, ii] = 0.
            #
            #     else:
            #         iim = ii-ng
            #         if (self.Kx[iim] != 0.) and (self.Ky[iim] != 0.):
            #             psi0[ii-ng, ii] = 1.j
            #             psi0[ii, ii] = -1.j
            #             self.phi0[ii, ii] = 1.
            #             self.phi0[ii-ng, ii] = -1.

            # # worked for normal incidence, Wood at (\pm 1, \pm 1) orders
            # ii = 0
            # psi0[5+ii, 5+ii] = 1.
            # psi0[5+ii+ng, 5+ii] = -1.
            # self.phi0[5+ii, 5+ii] = 0.
            # self.phi0[5+ii+ng, 5+ii] = 0.
            #
            # ii = 1
            # psi0[5+ii, 5+ii] = 0.
            # psi0[5+ii+ng, 5+ii] = 0.
            # self.phi0[5+ii, 5+ii] = 1.
            # self.phi0[5+ii+ng, 5+ii] = 1.
            #
            # ii = 2
            # psi0[5+ii, 5+ii] = 0.
            # psi0[5+ii+ng, 5+ii] = 0.
            # self.phi0[5+ii, 5+ii] = 1.
            # self.phi0[5+ii+ng, 5+ii] = 1.
            #
            # ii = 3
            # psi0[5+ii, 5+ii] = 0.
            # psi0[5+ii+ng, 5+ii] = 0.
            # self.phi0[5+ii, 5+ii] = 1.
            # self.phi0[5+ii+ng, 5+ii] = -1.
            #
            # ii = 0
            # psi0[5+ii+ng, 5+ii+ng] = 0.
            # psi0[5+ii, 5+ii+ng] = 0.
            # self.phi0[5+ii+ng, 5+ii+ng] = 1.
            # self.phi0[5+ii, 5+ii+ng] = -1.
            #
            # ii = 1
            # psi0[5+ii+ng, 5+ii+ng] = 1.
            # psi0[5+ii, 5+ii+ng] = 1.
            # self.phi0[5+ii+ng, 5+ii+ng] = 0.
            # self.phi0[5+ii, 5+ii+ng] = 0.
            #
            # ii = 2
            # psi0[5+ii+ng, 5+ii+ng] = 1.
            # psi0[5+ii, 5+ii+ng] = 1.
            # self.phi0[5+ii+ng, 5+ii+ng] = 0.
            # self.phi0[5+ii, 5+ii+ng] = 0.
            #
            # ii = 3
            # psi0[5+ii+ng, 5+ii+ng] = 1.
            # psi0[5+ii, 5+ii+ng] = -1.
            # self.phi0[5+ii+ng, 5+ii+ng] = 0.
            # self.phi0[5+ii, 5+ii+ng] = 0.

            self.psi0 = psi0

            # print('_calc_psi0', time.process_time() - t1)

    def _calc_im0(self):
        if self._num_g_ac:
            a0 = np.diag(2 * np.ones(2 * self._num_g_ac))
            b0 = np.zeros((2 * self._num_g_ac, 2 * self._num_g_ac))
            self.im0 = (a0, b0)

    def _calc_s_0(self):
        """calculate vacuum scattering matrix"""
        # todo: change into sparse to save memory
        if self._num_g_ac:
            # t1 = time.process_time()
            s11_0 = np.zeros((2 * self._num_g_ac, 2 * self._num_g_ac), dtype=complex)
            s12_0 = np.eye(2 * self._num_g_ac, dtype=complex)
            s21_0 = s12_0
            s22_0 = s11_0
            self.sm0 = (s11_0, s12_0, s21_0, s22_0)
            # print('_calc_s_0_3d', time.process_time() - t1)

    @property
    def uc_area(self) -> float:
        """Unit cell area"""
        return self._uc_area

    def _calc_uc_area(self):
        if self.latt_vec:
            if type(self.latt_vec) is not tuple:
                self._uc_area = self.latt_vec
            else:
                a1, a2 = self.latt_vec
                a1n, a2n = [la.norm(a) for a in [a1, a2]]
                if a1n != 0. and a2n != 0.:
                    self._uc_area = np.abs(np.cross(self.latt_vec[0], self.latt_vec[1]))
                elif a1n == 0.:  # 1D
                    self._uc_area = a2n
                elif a2n == 0.:
                    self._uc_area = a1n
                else:
                    raise Exception("Both lattice vectors have zero norms. Can't calculate unit cell area.")

    def _calc_angles(self):
        """
        calculate cos(vartheta), sin(vartheta), cos(phi), sin(phi) for all relevant orders, where vartheta = pi/2-theta
        This is used for setting incidence in high orders.
        """
        if self.ks:
            # t1 = time.process_time()

            idxa = np.array(self.idx_g)
            ksa = np.array(self.ks)
            k_pa = la.norm(ksa, axis=-1)  # norm of in-plane momentum
            i0 = (k_pa == 0.)
            ib = (k_pa != 0.)
            ii = (idxa[:, 0] == 0) & (idxa[:, 1] == 0)

            cphi = np.zeros(self._num_g_ac, complex)
            sphi = np.zeros(self._num_g_ac, complex)
            cphi[ib] = ksa[ib, 0] / k_pa[ib]
            sphi[ib] = ksa[ib, 1] / k_pa[ib]
            cphi[i0] = 1.
            sphi[i0] = 0.
            cphi[ii] = np.cos(self._phi)
            sphi[ii] = np.sin(self._phi)

            if self.inci_is_iso_nonvac:
                k = self.omega.real * self.ind_inci
            else:
                k = self.omega.real
            cthe = k_pa / k + 0j  # this is always positive
            # todo: when omega complex, k_parallel is still calc as real
            sthe = np.sqrt(1 - cthe**2 + 0j)
            cthe[ii] = np.cos(np.pi/2 - self._theta)
            sthe[ii] = np.sin(np.pi/2 - self._theta)

            if self.out_is_iso_nonvac:
                k = self.omega.real * self.ind_out
            else:
                k = self.omega.real
            cthe_bk = k_pa / k + 0j  # this is always positive
            # todo: when omega complex, k_parallel is still calc as real
            sthe_bk = np.sqrt(1 - cthe_bk**2 + 0j)
            cthe_bk[ii] = np.cos(np.pi/2 - self._theta)
            sthe_bk[ii] = np.sin(np.pi/2 - self._theta)

            self.cos_phis = list(cphi)
            self.sin_phis = list(sphi)
            self.cos_phis_bk = self.cos_phis.copy()
            self.sin_phis_bk = self.sin_phis.copy()
            self.cos_varthetas = list(cthe)
            self.sin_varthetas = list(sthe)
            self.cos_varthetas_bk = list(cthe)
            self.sin_varthetas_bk = list(sthe)


            # print('_calc_angles', time.process_time()-t1)
        # self._calc_ai_bo_3d()

    def _calc_ai_bo_3d(self):
        """calculate incident ai and bo amplitudes"""
        warn('this method is deprecated. It is moved to simulator', DeprecationWarning)

        # t1 = time.process_time()

        o = self.omega

        aibo = [self.ai, self.bo]
        for ii, (sa, pa, od) in enumerate([[self._s_amps, self._p_amps, self._incident_orders], [self._s_amps_bk, self._p_amps_bk, self._incident_orders_bk]]):
            if self._num_g_ac:
                ab = np.zeros(2 * self._num_g_ac) + 0j
                if (sa or pa) and od and self.idx_g and \
                        self.sin_phis and self.sin_varthetas and self.cos_phis and self.cos_varthetas:
                    # find the index of the input orders in the g list
                    idx = [i for order in od for i, j in enumerate(self.idx_g) if j == order]
                    for i, jj in enumerate(idx):
                    # for i in range(len(idx)):
                        if (jj in self.q0_0):
                            # todo: need to handle this and document it.
                            #  if user specify 90 degree incidence, this is activated
                            warn('You are specifying incidence in a channel that is parallel to the surface of the structure. \n In this case, only specific field configuration is allowed.')
                            ab[jj] = sa[i]
                            ab[jj + self._num_g_ac] = pa[i]
                        else:
                            s = sa[i]
                            p = pa[i]

                            # # original
                            # ab[i] = -s * self.sin_phis[i] + p * self.sin_varthetas[i] * self.cos_phis[i]  # e_x
                            # ab[i + self._num_g_ac] = s * self.cos_phis[i] + p * self.sin_varthetas[i] * self.sin_phis[i]  # e_y

                            # # original, corrected
                            # ab[jj] = -s * self.sin_phis[jj] + p * self.sin_varthetas[jj] * self.cos_phis[jj]  # e_x
                            # ab[jj + self._num_g_ac] = s * self.cos_phis[jj] + p * self.sin_varthetas[jj] * self.sin_phis[jj]  # e_y

                            # with new calc that removed convergence problem at Wood
                            ex = -s * self.sin_phis[jj] + p * self.sin_varthetas[jj] * self.cos_phis[jj]  # e_x
                            ey = s * self.cos_phis[jj] + p * self.sin_varthetas[jj] * self.sin_phis[jj]  # e_y
                            phi_2x2 = self.phi0_2x2s[:, :, jj]
                            phi_2x2_i = 1 / la.det(phi_2x2) * np.array([[phi_2x2[1, 1], -phi_2x2[0, 1]],
                                                                        [-phi_2x2[1, 0], phi_2x2[0, 0]]])
                            # v = la.solve(phi_2x2, np.array([ex, ey]))
                            v = phi_2x2_i @ np.array([ex, ey])
                            ab[jj] = v[0]
                            ab[jj + self._num_g_ac] = v[1]

                aibo[ii] = ab
        self.ai, self.bo = aibo
        # print('_calc_ai_bo_3d', time.process_time() - t1)




# -*- coding: utf-8 -*-

import numpy as np
import numpy.linalg as la
# from scipy import sparse as sps
# import scipy.fft as fft
from typing import Tuple, List, Union, Optional, Set
# import time
from warnings import warn
from inkstone.recipro import recipro
from inkstone.g_pts import g_pts
from inkstone.g_pts_1d import g_pts_1d
from inkstone.max_idx_diff import max_idx_diff
from inkstone.conv_mtx_idx import conv_mtx_idx_2d
from inkstone.params import Params

if __name__ == '__main__':
    import matplotlib.pyplot as plt

    lv = ((1., 0), (0.5, 1))
    num_g = 100
    omega = 0.5
    theta = 30.
    phi = 40.
    pr = Params(latt_vec=lv, num_g=num_g, omega=omega, theta=theta, phi=phi)



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
A photonic-crystal slab in vacuum. Permittivity 12.
Square lattice with period 1 in both x and y.
thickness 0.5
A circular hole of radius 0.2 in each unit cell.

 → y
↓x      ⋮
     ◯ ◯ ◯
...  ◯ ◯ ◯ ...  (top view)
     ◯ ◯ ◯
        ⋮
"""

import numpy as np
from inkstone import Inkstone

s = Inkstone()
s.lattice = ((1, 0), (0, 1))
s.num_g = 50

s.AddMaterial(name='di', epsilon=12)

s.AddLayer(name='in', thickness=0, material_background='vacuum')
s.AddLayer(name='slab', thickness=0.5, material_background='di')
s.AddLayerCopy(name='out', original_layer='in', thickness=0)

s.AddPatternDisk(layer='slab', pattern_name='disk', material='vacuum', radius=0.2)

# Incident wave
s.SetExcitation(theta=0, phi=0, s_amplitude=1, p_amplitude=0)

flux_in = []
flux_out = []
freq = np.linspace(0.25, 0.45, 201)
for i in freq:
    print('Frequency: {:g}'.format(i))

    s.SetFrequency(i)

    flux_in.append(s.GetPowerFlux('in'))
    flux_out.append(s.GetPowerFlux('out'))

incident = np.array([a[0] for a in flux_in])
reflection = -np.array([a[1] for a in flux_in]) / incident
transmission = np.array([a[0] for a in flux_out]) / incident

#%% plotting
from matplotlib import pyplot as plt
plt.plot(freq, transmission)
plt.xlabel('frequency')
plt.ylabel('transmission')

